@@ -1,21 +1,21 @@
- /*
-  The oRTP library is an RTP (Realtime Transport Protocol - rfc3550) stack.
-  Copyright (C) 2001  Simon MORLAT simon.morlat@linphone.org
-
-  This library is free software; you can redistribute it and/or
-  modify it under the terms of the GNU Lesser General Public
-  License as published by the Free Software Foundation; either
-  version 2.1 of the License, or (at your option) any later version.
-
-  This library is distributed in the hope that it will be useful,
-  but WITHOUT ANY WARRANTY; without even the implied warranty of
-  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
-  Lesser General Public License for more details.
-
-  You should have received a copy of the GNU Lesser General Public
-  License along with this library; if not, write to the Free Software
-  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
-*/
+/*
+ * Copyright (c) 2010-2019 Belledonne Communications SARL.
+ *
+ * This file is part of oRTP.
+ *
+ * This program is free software: you can redistribute it and/or modify
+ * it under the terms of the GNU General Public License as published by
+ * the Free Software Foundation, either version 3 of the License, or
+ * (at your option) any later version.
+ *
+ * This program is distributed in the hope that it will be useful,
+ * but WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+ * GNU General Public License for more details.
+ *
+ * You should have received a copy of the GNU General Public License
+ * along with this program. If not, see <http://www.gnu.org/licenses/>.
+ */
 
 /**
  * \file rtpsession.h
@@ -31,6 +31,8 @@
 #define RTPSESSION_H
 
 
+#include <bctoolbox/list.h>
+
 #include <ortp/port.h>
 #include <ortp/rtp.h>
 #include <ortp/payloadtype.h>
@@ -38,12 +40,15 @@
 #include <ortp/sessionset.h>
 #include <ortp/rtcp.h>
 #include <ortp/str_utils.h>
+#include <ortp/utils.h>
 #include <ortp/rtpsignaltable.h>
 #include <ortp/event.h>
-
+#include <ortp/fecstream.h>
 
 #define ORTP_AVPF_FEATURE_NONE 0
 #define ORTP_AVPF_FEATURE_TMMBR (1 << 0)
+#define ORTP_AVPF_FEATURE_GENERIC_NACK (1 << 1)
+#define ORTP_AVPF_FEATURE_IMMEDIATE_NACK (1 << 2)
 
 
 typedef enum {
@@ -53,35 +58,56 @@ typedef enum {
 } RtpSessionMode;
 
 
+typedef enum _OrtpJitterBufferAlgorithm {
+	OrtpJitterBufferBasic,
+	OrtpJitterBufferRecursiveLeastSquare,
+} OrtpJitterBufferAlgorithm;
+
 /*! Jitter buffer parameters
 */
 typedef struct _JBParameters{
-	int min_size; /**< in milliseconds*/
-	int nom_size; /**< idem */
-	int max_size; /**< idem */
-	bool_t adaptive;
-	bool_t pad[3];
-	int max_packets; /**< max number of packets allowed to be queued in the jitter buffer */
+	int min_size; /*(adaptive=TRUE only) maximum dynamic delay to be added to incoming packets (ms) */
+	int nom_size; /*(adaptive=TRUE only) initial dynamic delay to be added to incoming packets (ms) */
+	int max_size; /*(adaptive=TRUE only) minimum dynamic delay to be added to incoming packets (ms) */
+	bool_t adaptive; /*either a dynamic buffer should be used or not to compensate bursts */
+	bool_t enabled; /*whether jitter buffer is enabled*/
+	bool_t pad[2]; /*(dev only) alignment pad: insert your bool_t here*/
+	int max_packets; /**max number of packets allowed to be queued in the jitter buffer */
+	OrtpJitterBufferAlgorithm buffer_algorithm;
+	int refresh_ms; /* (adaptive=TRUE only) dynamic buffer size update frequency (ms) */
+	int ramp_threshold; /*(adaptive=TRUE, algo=RLS only) Percentage in [0;100] threshold between current jitter and previous jitter to enable smooth ramp*/
+	int ramp_step_ms; /*(adaptive=TRUE, algo=RLS only) In smooth ramp, how much we should reduce jitter size on each step*/
+	int ramp_refresh_ms; /*(adaptive=TRUE, algo=RLS only) In smooth ramp, frequency of step*/
 } JBParameters;
 
 typedef struct _JitterControl
 {
-	unsigned int count;
-	int jitt_comp;   /* the user jitt_comp in miliseconds*/
-	int jitt_comp_ts; /* the jitt_comp converted in rtp time (same unit as timestamp) */
+	JBParameters params;
+	unsigned int count; /* number of packets handled in jitter_control_new_packet. Used internally only. */
+	int jitt_comp_ts; /* the nominal jitter buffer size converted in rtp time (same unit as timestamp) */
 	int adapt_jitt_comp_ts;
-	int64_t slide;
-	int64_t prev_slide;
+	int32_t clock_offset_ts; /*offset difference between local and distant clock, in timestamp units*/
+	int32_t prev_clock_offset_ts;
+	int32_t olddiff;
 	float jitter;
-	int olddiff;
 	float inter_jitter;	/* interarrival jitter as defined in the RFC */
+	float jitter_buffer_mean_size; /*effective size (fullness) of jitter buffer*/
 	int corrective_step;
 	int corrective_slide;
 	uint64_t cum_jitter_buffer_size; /*in timestamp units*/
 	unsigned int cum_jitter_buffer_count; /*used for computation of jitter buffer size*/
 	int clock_rate;
-	bool_t adaptive;
-	bool_t enabled;
+	uint32_t adapt_refresh_prev_ts; /*last time we refreshed the buffer*/
+	OrtpExtremum max_ts_deviation; /*maximum difference between packet and expected timestamps */
+	OrtpKalmanRLS kalman_rls;
+	double capped_clock_ratio;
+	uint32_t last_log_ts;
+	uint32_t local_ts_start;
+	uint32_t remote_ts_start;
+	uint32_t diverged_start_ts;
+	bool_t is_diverging;
+	bool_t jb_size_updated;
+	bool_t pad[2];
 } JitterControl;
 
 typedef struct _WaitPoint
@@ -96,12 +122,13 @@ typedef struct _RtpTransportModifier
 {
 	void *data;
 	struct _RtpSession *session;//<back pointer to the owning session, set by oRTP
+	struct _RtpTransport *transport;//<back point to the owning transport, set by oRTP
 	int  (*t_process_on_send)(struct _RtpTransportModifier *t, mblk_t *msg);
 	int  (*t_process_on_receive)(struct _RtpTransportModifier *t, mblk_t *msg);
-	void  (*t_process_on_schedule)(struct _RtpTransportModifier *t); /*invoked each time rtp_session_recvm even is no message are available*/
+	void  (*t_process_on_schedule)(struct _RtpTransportModifier *t); /*invoked each time rtp_session_recvm is called even is no message are available*/
 	/**
-	 * Mandatory callback responsible of freeing the #RtpTransportModifierAND the pointer.
-	 * @param[in] transport #RtpTransportModifier object to free.
+	 * Mandatory callback responsible of freeing the #_RtpTransportModifier AND the pointer.
+	 * @param[in] transport #_RtpTransportModifier object to free.
 	 */
 	void  (*t_destroy)(struct _RtpTransportModifier *transport);
 } RtpTransportModifier;
@@ -113,10 +140,10 @@ typedef struct _RtpTransport
 	ortp_socket_t (*t_getsocket)(struct _RtpTransport *t);
 	int  (*t_sendto)(struct _RtpTransport *t, mblk_t *msg , int flags, const struct sockaddr *to, socklen_t tolen);
 	int  (*t_recvfrom)(struct _RtpTransport *t, mblk_t *msg, int flags, struct sockaddr *from, socklen_t *fromlen);
-	void  (*t_close)(struct _RtpTransport *transport, void *userData);
+	void  (*t_close)(struct _RtpTransport *transport);
 	/**
-	 * Mandatory callback responsible of freeing the #RtpTransport object AND the pointer.
-	 * @param[in] transport #RtpTransport object to free.
+	 * Mandatory callback responsible of freeing the #_RtpTransport object AND the pointer.
+	 * @param[in] transport #_RtpTransport object to free.
 	 */
 	void  (*t_destroy)(struct _RtpTransport *transport);
 }  RtpTransport;
@@ -125,8 +152,8 @@ typedef enum _OrtpNetworkSimulatorMode{
 	OrtpNetworkSimulatorInvalid=-1,
 	OrtpNetworkSimulatorInbound,/**<simulation is applied when receiving packets*/
 	OrtpNetworkSimulatorOutbound, /**<simulation is applied to sent packets*/
-	OrtpNetworkSimulatorOutboundControlled /**<simulation is applied to sent packets according to sent timestamp 
-				set in the timestamp field of mblk_t, which is definied only with -DORTP_TIMESTAMP */
+	OrtpNetworkSimulatorOutboundControlled /**<simulation is applied to sent packets according to sent timestamps
+				set in the timestamps field of mblk_t, which is defined only with -DORTP_TIMESTAMP */
 }OrtpNetworkSimulatorMode;
 
 /**
@@ -134,14 +161,18 @@ typedef enum _OrtpNetworkSimulatorMode{
 **/
 typedef struct _OrtpNetworkSimulatorParams{
 	int enabled; /**<Whether simulation is enabled or off.*/
-	float max_bandwidth; /**<IP bandwidth, in bit/s*/
+	float max_bandwidth; /**<IP bandwidth, in bit/s.
+						This limitation is applied after loss are simulated, so incoming bandwidth
+						is NOT socket bandwidth, but after-loss-simulation bandwidth e.g with 50% loss, the bandwidth will be 50% reduced*/
 	int max_buffer_size; /**<Max number of bit buffered before being discarded*/
 	float loss_rate; /**<Percentage of lost packets*/
 	uint32_t latency; /**<Packet transmission delay, in ms*/
-	float consecutive_loss_probability;/**< a probablity of having a subsequent loss after a loss occured, in a 0-1 range. Useful to simulate burst of lost packets*/
+	float consecutive_loss_probability;/**< a probability of having a subsequent loss after a loss occurred, in a 0-1 range. Useful to simulate burst of lost packets*/
 	float jitter_burst_density; /**<density of gap/bursts events. A value of 1 means one gap/burst per second approximately*/
-	float jitter_strength; /**<percentage of max_bandwidth artifically consumed during bursts events*/
-	OrtpNetworkSimulatorMode mode; /**<whether simulation is applied to inboud or outbound stream.*/
+	float jitter_strength; /**<percentage of max_bandwidth artificially consumed during bursts events*/
+	bool_t rtp_only; /**True for only RTP packet loss, False for both RTP and RTCP */
+	bool_t pad[3];
+	OrtpNetworkSimulatorMode mode; /**<whether simulation is applied to inbound or outbound stream.*/
 }OrtpNetworkSimulatorParams;
 
 typedef struct _OrtpNetworkSimulatorCtx{
@@ -182,6 +213,11 @@ typedef struct OrtpRtcpSendAlgorithm {
 	bool_t tmmbn_scheduled;
 } OrtpRtcpSendAlgorithm;
 
+typedef struct OrtpRtcpFbConfiguration {
+	bool_t generic_nack_enabled;
+	bool_t tmmbr_enabled;
+} OrtpRtcpFbConfiguration;
+
 #define ORTP_RTCP_XR_UNAVAILABLE_PARAMETER 127
 
 typedef enum {
@@ -190,10 +226,10 @@ typedef enum {
 	OrtpRtcpXrEnhancedPlc
 } OrtpRtcpXrPlcStatus;
 
-typedef OrtpRtcpXrPlcStatus (*OrtpRtcpXrPlcCallback)(unsigned long userdata);
-typedef int (*OrtpRtcpXrSignalLevelCallback)(unsigned long userdata);
-typedef int (*OrtpRtcpXrNoiseLevelCallback)(unsigned long userdata);
-typedef float (*OrtpRtcpXrAverageQualityIndicatorCallback)(unsigned long userdata);
+typedef OrtpRtcpXrPlcStatus (*OrtpRtcpXrPlcCallback)(void *userdata);
+typedef int (*OrtpRtcpXrSignalLevelCallback)(void *userdata);
+typedef int (*OrtpRtcpXrNoiseLevelCallback)(void *userdata);
+typedef float (*OrtpRtcpXrAverageQualityIndicatorCallback)(void *userdata);
 
 typedef struct OrtpRtcpXrMediaCallbacks {
 	OrtpRtcpXrPlcCallback plc;
@@ -201,7 +237,7 @@ typedef struct OrtpRtcpXrMediaCallbacks {
 	OrtpRtcpXrNoiseLevelCallback noise_level;
 	OrtpRtcpXrAverageQualityIndicatorCallback average_qi;
 	OrtpRtcpXrAverageQualityIndicatorCallback average_lq_qi;
-	unsigned long userdata;
+	void *userdata;
 } OrtpRtcpXrMediaCallbacks;
 
 typedef enum {
@@ -211,6 +247,7 @@ typedef enum {
 } OrtpRtcpXrRcvrRttMode;
 
 typedef enum {
+	OrtpRtcpXrStatSummaryNone = 0,
 	OrtpRtcpXrStatSummaryLoss = (1 << 7),
 	OrtpRtcpXrStatSummaryDup = (1 << 6),
 	OrtpRtcpXrStatSummaryJitt = (1 << 5),
@@ -222,6 +259,7 @@ typedef struct OrtpRtcpXrConfiguration {
 	bool_t enabled;
 	bool_t stat_summary_enabled;
 	bool_t voip_metrics_enabled;
+	bool_t pad;
 	OrtpRtcpXrRcvrRttMode rcvr_rtt_mode;
 	int rcvr_rtt_max_size;
 	OrtpRtcpXrStatSummaryFlag stat_summary_flags;
@@ -270,21 +308,25 @@ typedef struct _OrtpStream {
 	struct sockaddr_storage rem_addr;
 	socklen_t loc_addrlen;
 	struct sockaddr_storage loc_addr;
+	socklen_t used_loc_addrlen;
+	struct sockaddr_storage used_loc_addr; /*Address used to redirect packets from this source*/
 	struct _RtpTransport *tr;
-	mblk_t *cached_mp;
+	OrtpBwEstimator recv_bw_estimator;
 	struct timeval send_bw_start; /* used for bandwidth estimation */
 	struct timeval recv_bw_start; /* used for bandwidth estimation */
 	unsigned int sent_bytes; /* used for bandwidth estimation */
 	unsigned int recv_bytes; /* used for bandwidth estimation */
 	float upload_bw;
 	float download_bw;
-	OList *aux_destinations; /*list of OrtpAddress */
+	float average_upload_bw;
+	float average_download_bw;
+	bctbx_list_t *aux_destinations; /*list of OrtpAddress */
+	msgb_allocator_t allocator;
 } OrtpStream;
 
 typedef struct _RtpStream
 {
 	OrtpStream gs;
-	int max_rq_size;
 	int time_jump;
 	uint32_t ts_jump;
 	queue_t rq;
@@ -295,7 +337,8 @@ typedef struct _RtpStream
 	uint32_t snd_time_offset;/*the scheduler time when the application send its first timestamp*/
 	uint32_t snd_ts_offset;	/* the first application timestamp sent by the application */
 	uint32_t snd_rand_offset;	/* a random number added to the user offset to make the stream timestamp*/
-	uint32_t snd_last_ts;	/* the last stream timestamp sended */
+	uint32_t snd_last_ts;	/* the last stream timestamp sent */
+	uint16_t snd_last_nack;	/* the last nack sent when in immediate mode */
 	uint32_t rcv_time_offset; /*the scheduler time when the application ask for its first timestamp*/
 	uint32_t rcv_ts_offset;  /* the first stream timestamp */
 	uint32_t rcv_query_ts_offset;	/* the first user timestamp asked by the application */
@@ -312,13 +355,19 @@ typedef struct _RtpStream
 	uint16_t snd_seq; /* send sequence number */
 	uint32_t last_rtcp_packet_count; /*the sender's octet count in the last sent RTCP SR*/
 	uint32_t sent_payload_bytes; /*used for RTCP sender reports*/
-	rtp_stats_t stats;
 	int recv_errno;
 	int send_errno;
 	int snd_socket_size;
 	int rcv_socket_size;
 	int ssrc_changed_thres;
 	jitter_stats_t jitter_stats;
+	struct _OrtpCongestionDetector *congdetect;
+	struct _OrtpVideoBandwidthEstimator *video_bw_estimator;
+	ortp_thread_t win_t;
+	volatile bool_t is_win_thread_running;
+	ortp_mutex_t winthread_lock;
+	queue_t winrq;
+	ortp_mutex_t winrq_lock;
 }RtpStream;
 
 typedef struct _RtcpStream
@@ -336,7 +385,6 @@ typedef struct _RtcpStream
 
 typedef struct _RtpSession RtpSession;
 
-
 /**
  * An object representing a bi-directional RTP session.
  * It holds sockets, jitter buffer, various counters (timestamp, sequence numbers...)
@@ -347,6 +395,7 @@ typedef struct _RtpSession RtpSession;
 **/
 struct _RtpSession
 {
+	ortp_mutex_t main_mutex; /* To protect data that can be accessed simultaneously by a control thread and the real-time thread in charge of sending/receiving. */
 	RtpSession *next;	/* next RtpSession, when the session are enqueued by the scheduler */
 	int mask_pos;	/* the position in the scheduler mask of RtpSession : do not move this field: it is part of the ABI since the session_set macros use it*/
 	struct {
@@ -359,7 +408,7 @@ struct _RtpSession
 	int inc_same_ssrc_count;
 	int hw_recv_pt; /* recv payload type before jitter buffer */
 	int recv_buf_size;
-	int target_upload_bandwidth; /* Target upload bandwidth at nework layer (with IP and UDP headers) in bits/s */
+	int target_upload_bandwidth; /* Target upload bandwidth at network layer (with IP and UDP headers) in bits/s */
 	RtpSignalTable on_ssrc_changed;
 	RtpSignalTable on_payload_type_changed;
 	RtpSignalTable on_telephone_event_packet;
@@ -367,9 +416,8 @@ struct _RtpSession
 	RtpSignalTable on_timestamp_jump;
 	RtpSignalTable on_network_error;
 	RtpSignalTable on_rtcp_bye;
-	struct _OList *signal_tables;
-	struct _OList *eventqs;
-	msgb_allocator_t allocator;
+	bctbx_list_t *signal_tables;
+	bctbx_list_t *eventqs;
 	RtpStream rtp;
 	RtcpStream rtcp;
 	OrtpRtcpXrStats rtcp_xr_stats;
@@ -380,30 +428,63 @@ struct _RtpSession
 	int multicast_ttl;
 	int multicast_loopback;
 	float duplication_ratio; /* Number of times a packet should be duplicated */
-	float duplication_left ; /* Remainder of the duplication ratio */
+	float duplication_left ; /* Remainder of the duplication ratio, internal use */
 	void * user_data;
 	/* FIXME: Should be a table for all session participants. */
 	struct timeval last_recv_time; /* Time of receiving the RTP/RTCP packet. */
 	mblk_t *pending;
 	/* telephony events extension */
+	int tev_send_pt; /*telephone event to be used for sending*/
 	mblk_t *current_tev;		/* the pending telephony events */
 	mblk_t *minimal_sdes;
 	mblk_t *full_sdes;
 	queue_t contributing_sources;
-	int64_t lost_packets_test_vector;
+	int lost_packets_test_vector;
 	unsigned int interarrival_jitter_test_vector;
 	unsigned int delay_test_vector;
 	float rtt;/*last round trip delay calculated*/
 	int cum_loss;
 	OrtpNetworkSimulatorCtx *net_sim_ctx;
+	RtpSession *spliced_session; /*a RtpSession that will retransmit everything received on this session*/
+	rtp_stats_t stats;
+	bctbx_list_t *recv_addr_map;
+	uint32_t send_ts_offset; /*additional offset to add when sending packets */
 	bool_t symmetric_rtp;
 	bool_t permissive; /*use the permissive algorithm*/
 	bool_t use_connect; /* use connect() on the socket */
 	bool_t ssrc_set;
+
 	bool_t reuseaddr; /*setsockopt SO_REUSEADDR */
+	bool_t rtcp_mux;
 	unsigned char avpf_features; /**< A bitmask of ORTP_AVPF_FEATURE_* macros. */
+	bool_t use_pktinfo;
+
+	bool_t is_spliced;
+	bool_t congestion_detector_enabled;
+	bool_t video_bandwidth_estimator_enabled;
+	bool_t is_primary;  /* tells if this session is the primary of the rtp bundle */
+	
+	bool_t warn_non_working_pkt_info;
+
+	/* bundle mode */
+	struct _RtpBundle *bundle; /* back pointer to the rtp bundle object */
+	queue_t bundleq;
+	ortp_mutex_t bundleq_lock;
+
+    /* fec option */
+    FecStream *fec_stream;
 };
 
+/**
+ * Structure describing the video bandwidth estimator parameters
+**/
+typedef struct _OrtpVideoBandwidthEstimatorParams {
+	int enabled; /**<Whether estimator is enabled or off.*/
+	unsigned int packet_count_min; /** minimum number of packets with the same sent timestamp to be processed continuously before being used */
+	unsigned int packets_size_max; /** number of packets needed to compute the available video bandwidth */
+	unsigned int trust_percentage; /** percentage for which the chosen bandwidth value in all available will be inferior */
+} OrtpVideoBandwidthEstimatorParams;
+
 
 
 
@@ -415,7 +496,7 @@ extern "C"
 ORTP_PUBLIC const char *ortp_network_simulator_mode_to_string(OrtpNetworkSimulatorMode mode);
 ORTP_PUBLIC OrtpNetworkSimulatorMode ortp_network_simulator_mode_from_string(const char *str);
 
-	
+
 /* public API */
 ORTP_PUBLIC RtpSession *rtp_session_new(int mode);
 ORTP_PUBLIC void rtp_session_set_scheduling_mode(RtpSession *session, int yesno);
@@ -442,6 +523,16 @@ ORTP_PUBLIC bool_t rtp_session_jitter_buffer_enabled(const RtpSession *session);
 ORTP_PUBLIC void rtp_session_set_jitter_buffer_params(RtpSession *session, const JBParameters *par);
 ORTP_PUBLIC void rtp_session_get_jitter_buffer_params(RtpSession *session, JBParameters *par);
 
+/**
+ * Set an additional timestamps offset for outgoing stream..
+ * @param s		a rtp session freshly created.
+ * @param offset		a timestamp offset value
+ *
+**/
+ORTP_PUBLIC void rtp_session_set_send_ts_offset(RtpSession *s, uint32_t offset);
+ORTP_PUBLIC uint32_t rtp_session_get_send_ts_offset(RtpSession *s);
+
+
 /*deprecated jitter control functions*/
 ORTP_PUBLIC void rtp_session_set_jitter_compensation(RtpSession *session, int milisec);
 ORTP_PUBLIC void rtp_session_enable_adaptive_jitter_compensation(RtpSession *session, bool_t val);
@@ -468,6 +559,7 @@ ORTP_PUBLIC void rtp_session_get_transports(const RtpSession *session, RtpTransp
 /*those methods are provided for people who wants to send non-RTP messages using the RTP/RTCP sockets */
 ORTP_PUBLIC ortp_socket_t rtp_session_get_rtp_socket(const RtpSession *session);
 ORTP_PUBLIC ortp_socket_t rtp_session_get_rtcp_socket(const RtpSession *session);
+ORTP_PUBLIC void rtp_session_refresh_sockets(RtpSession *session);
 
 
 /* QOS / DSCP */
@@ -493,19 +585,39 @@ ORTP_PUBLIC int rtp_session_get_send_payload_type(const RtpSession *session);
 ORTP_PUBLIC int rtp_session_get_recv_payload_type(const RtpSession *session);
 ORTP_PUBLIC int rtp_session_set_recv_payload_type(RtpSession *session, int pt);
 
+ORTP_PUBLIC int rtp_session_set_send_telephone_event_payload_type(RtpSession *session, int paytype);
+
 ORTP_PUBLIC int rtp_session_set_payload_type(RtpSession *session, int pt);
 
 ORTP_PUBLIC void rtp_session_set_symmetric_rtp (RtpSession * session, bool_t yesno);
 
 ORTP_PUBLIC bool_t rtp_session_get_symmetric_rtp (const RtpSession * session);
 
+ORTP_PUBLIC void rtp_session_enable_rtcp_mux(RtpSession *session, bool_t yesno);
+
+ORTP_PUBLIC bool_t rtp_session_rtcp_mux_enabled(RtpSession *session);
+
 ORTP_PUBLIC void rtp_session_set_connected_mode(RtpSession *session, bool_t yesno);
 
 ORTP_PUBLIC void rtp_session_enable_rtcp(RtpSession *session, bool_t yesno);
+/*
+ * rtcp status
+ * @return TRUE if rtcp is enabled for this session
+ */
+ORTP_PUBLIC bool_t rtp_session_rtcp_enabled(const RtpSession *session);
 
 ORTP_PUBLIC void rtp_session_set_rtcp_report_interval(RtpSession *session, int value_ms);
 
+/**
+ * Define the bandwidth available for RTCP streams based on the upload bandwidth
+ * targeted by the application (in bits/s). RTCP streams would not take more than
+ * a few percents of the limit bandwidth (around 5%).
+ *
+ * @param session a rtp session
+ * @param target_bandwidth bandwidth limit in bits/s
+ */
 ORTP_PUBLIC void rtp_session_set_target_upload_bandwidth(RtpSession *session, int target_bandwidth);
+ORTP_PUBLIC int rtp_session_get_target_upload_bandwidth(RtpSession *session);
 
 ORTP_PUBLIC void rtp_session_configure_rtcp_xr(RtpSession *session, const OrtpRtcpXrConfiguration *config);
 ORTP_PUBLIC void rtp_session_set_rtcp_xr_media_callbacks(RtpSession *session, const OrtpRtcpXrMediaCallbacks *cbs);
@@ -514,15 +626,21 @@ ORTP_PUBLIC void rtp_session_set_ssrc_changed_threshold(RtpSession *session, int
 
 /*low level recv and send functions */
 ORTP_PUBLIC mblk_t * rtp_session_recvm_with_ts (RtpSession * session, uint32_t user_ts);
-ORTP_PUBLIC mblk_t * rtp_session_create_packet(RtpSession *session,int header_size, const uint8_t *payload, int payload_size);
-ORTP_PUBLIC mblk_t * rtp_session_create_packet_raw(const uint8_t *packet, int packet_size);
-ORTP_PUBLIC mblk_t * rtp_session_create_packet_with_data(RtpSession *session, uint8_t *payload, int payload_size, void (*freefn)(void*));
-ORTP_PUBLIC mblk_t * rtp_session_create_packet_in_place(RtpSession *session,uint8_t *buffer, int size, void (*freefn)(void*) );
+ORTP_PUBLIC mblk_t * rtp_session_create_packet(RtpSession *session, size_t header_size, const uint8_t *payload, size_t payload_size);
+ORTP_PUBLIC mblk_t * rtp_session_create_packet_raw(const uint8_t *packet, size_t packet_size);
+ORTP_PUBLIC mblk_t * rtp_session_create_packet_with_data(RtpSession *session, uint8_t *payload, size_t payload_size, void (*freefn)(void*));
+ORTP_PUBLIC mblk_t * rtp_session_create_packet_in_place(RtpSession *session,uint8_t *buffer, size_t size, void (*freefn)(void*) );
+ORTP_PUBLIC mblk_t * rtp_session_create_packet_with_mixer_to_client_audio_level(RtpSession *session, size_t header_size, int mtc_extension_id, size_t audio_levels_size, rtp_audio_level_t *audio_levels);
 ORTP_PUBLIC int rtp_session_sendm_with_ts (RtpSession * session, mblk_t *mp, uint32_t userts);
+ORTP_PUBLIC int rtp_session_sendto(RtpSession *session, bool_t is_rtp, mblk_t *m, int flags, const struct sockaddr *destaddr, socklen_t destlen);
+ORTP_PUBLIC int rtp_session_recvfrom(RtpSession *session, bool_t is_rtp, mblk_t *m, int flags, struct sockaddr *from, socklen_t *fromlen);
 /* high level recv and send functions */
 ORTP_PUBLIC int rtp_session_recv_with_ts(RtpSession *session, uint8_t *buffer, int len, uint32_t ts, int *have_more);
 ORTP_PUBLIC int rtp_session_send_with_ts(RtpSession *session, const uint8_t *buffer, int len, uint32_t userts);
 
+/* Specific function called to reset the winrq queue and if called on windows to stop the async reception thread */
+ORTP_PUBLIC void rtp_session_reset_recvfrom(RtpSession *session);
+
 /* event API*/
 ORTP_PUBLIC void rtp_session_register_event_queue(RtpSession *session, OrtpEvQueue *q);
 ORTP_PUBLIC void rtp_session_unregister_event_queue(RtpSession *session, OrtpEvQueue *q);
@@ -538,7 +656,21 @@ ORTP_PUBLIC float rtp_session_get_rtp_recv_bandwidth(RtpSession *session);
 ORTP_PUBLIC float rtp_session_get_rtcp_send_bandwidth(RtpSession *session);
 ORTP_PUBLIC float rtp_session_get_rtcp_recv_bandwidth(RtpSession *session);
 
+ORTP_PUBLIC float rtp_session_get_send_bandwidth_smooth(RtpSession *session);
+ORTP_PUBLIC float rtp_session_get_recv_bandwidth_smooth(RtpSession *session);
+
 ORTP_PUBLIC void rtp_session_send_rtcp_APP(RtpSession *session, uint8_t subtype, const char *name, const uint8_t *data, int datalen);
+/**
+ *	Send the rtcp datagram \a packet to the destination set by rtp_session_set_remote_addr()
+ *  The packet (\a packet) is freed once it is sent.
+ *
+ * @param session a rtp session.
+ * @param m a rtcp packet presented as a mblk_t.
+ * @return the number of bytes sent over the network.
+ **/
+
+ORTP_PUBLIC	int rtp_session_rtcp_sendm_raw(RtpSession * session, mblk_t * m);
+
 
 ORTP_PUBLIC uint32_t rtp_session_get_current_send_ts(RtpSession *session);
 ORTP_PUBLIC uint32_t rtp_session_get_current_recv_ts(RtpSession *session);
@@ -592,8 +724,10 @@ ORTP_PUBLIC float rtp_session_get_round_trip_propagation(RtpSession *session);
 
 
 ORTP_PUBLIC void rtp_session_enable_network_simulation(RtpSession *session, const OrtpNetworkSimulatorParams *params);
+ORTP_PUBLIC void rtp_session_enable_congestion_detection(RtpSession *session, bool_t enabled);
+ORTP_PUBLIC void rtp_session_enable_video_bandwidth_estimator(RtpSession *session, const OrtpVideoBandwidthEstimatorParams *params);
 
-ORTP_PUBLIC void rtp_session_rtcp_set_lost_packet_value( RtpSession *session, const int64_t value );
+ORTP_PUBLIC void rtp_session_rtcp_set_lost_packet_value( RtpSession *session, const int value );
 ORTP_PUBLIC void rtp_session_rtcp_set_jitter_value(RtpSession *session, const unsigned int value );
 ORTP_PUBLIC void rtp_session_rtcp_set_delay_value(RtpSession *session, const unsigned int value );
 ORTP_PUBLIC mblk_t * rtp_session_pick_with_cseq (RtpSession * session, const uint16_t sequence_number);
@@ -612,6 +746,7 @@ ORTP_PUBLIC void rtp_session_enable_avpf_feature(RtpSession *session, unsigned c
 ORTP_PUBLIC uint16_t rtp_session_get_avpf_rr_interval(RtpSession *session);
 ORTP_PUBLIC bool_t rtp_session_rtcp_psfb_scheduled(RtpSession *session, rtcp_psfb_type_t type);
 ORTP_PUBLIC bool_t rtp_session_rtcp_rtpfb_scheduled(RtpSession *session, rtcp_rtpfb_type_t type);
+ORTP_PUBLIC void rtp_session_send_rtcp_fb_generic_nack(RtpSession *session, uint16_t pid, uint16_t blp);
 ORTP_PUBLIC void rtp_session_send_rtcp_fb_pli(RtpSession *session);
 ORTP_PUBLIC void rtp_session_send_rtcp_fb_fir(RtpSession *session);
 ORTP_PUBLIC void rtp_session_send_rtcp_fb_sli(RtpSession *session, uint16_t first, uint16_t number, uint8_t picture_id);
@@ -627,40 +762,64 @@ ORTP_PUBLIC void rtp_session_init(RtpSession *session, int mode);
 ORTP_PUBLIC void rtp_session_uninit(RtpSession *session);
 ORTP_PUBLIC void rtp_session_dispatch_event(RtpSession *session, OrtpEvent *ev);
 
+
 ORTP_PUBLIC void rtp_session_set_reuseaddr(RtpSession *session, bool_t yes);
 
+ORTP_PUBLIC int meta_rtp_transport_sendto(RtpTransport *t, mblk_t *msg , int flags, const struct sockaddr *to, socklen_t tolen);
 
-ORTP_PUBLIC int meta_rtp_transport_modifier_inject_packet(const RtpTransport *t, RtpTransportModifier *tpm, mblk_t *msg , int flags);
-/**
- * #RtpTransport object which can handle multiples security protocols. You can for instance use this object
- * to use both sRTP and tunnel transporter. #mblk_t messages received and sent from the endpoint
- * will pass through the list of modifiers given. First modifier in list will be first to modify the message
- * in send mode and last in receive mode.
- * @param[in] t #RtpTransport object that will be generated.
- * @param[in] is_rtp Whether this object will be used for RTP packets or not.
- * @param[in] endpoint #RtpTransport object in charge of sending/receiving packets. If NULL, it will use standards sendto and recvfrom functions.
- * @param[in] modifiers_count number of #RtpModifier object given in the variadic list. Must be 0 if none are given.
- * @return 0 if successful, -1 otherwise
-**/
-ORTP_PUBLIC int meta_rtp_transport_modifier_inject_packet_to(const RtpTransport *t, RtpTransportModifier *tpm, mblk_t *msg , int flags,const struct sockaddr *to, socklen_t tolen) ;
+ORTP_PUBLIC int meta_rtp_transport_modifier_inject_packet_to_send(RtpTransport *t, RtpTransportModifier *tpm, mblk_t *msg, int flags);
+ORTP_PUBLIC int meta_rtp_transport_modifier_inject_packet_to_send_to(RtpTransport *t, RtpTransportModifier *tpm, mblk_t *msg, int flags, const struct sockaddr *to, socklen_t tolen);
+ORTP_PUBLIC int meta_rtp_transport_modifier_inject_packet_to_recv(RtpTransport *t, RtpTransportModifier *tpm, mblk_t *msg, int flags);
 
 /**
  * get endpoint if any
- * @param[in] t #RtpTransport object.
- * @return #rtpEndpoint
+ * @param[in] transport RtpTransport object.
+ * @return #_RtpTransport
  *
  * */
 ORTP_PUBLIC RtpTransport* meta_rtp_transport_get_endpoint(const RtpTransport *transport);
 /**
  * set endpoint
- * @param[in] t #RtpTransport object.
- * @param[in] t #RtpEndpoint.
+ * @param[in] transport RtpTransport object.
+ * @param[in] endpoint RtpEndpoint.
  *
  * */
 ORTP_PUBLIC void meta_rtp_transport_set_endpoint(RtpTransport *transport,RtpTransport *endpoint);
 
 ORTP_PUBLIC void meta_rtp_transport_destroy(RtpTransport *tp);
 ORTP_PUBLIC void meta_rtp_transport_append_modifier(RtpTransport *tp,RtpTransportModifier *tpm);
+ORTP_PUBLIC void meta_rtp_transport_prepend_modifier(RtpTransport *tp,RtpTransportModifier *tpm);
+ORTP_PUBLIC void meta_rtp_transport_remove_modifier(RtpTransport *tp, RtpTransportModifier *tpm);
+
+ORTP_PUBLIC int rtp_session_splice(RtpSession *session, RtpSession *to_session);
+ORTP_PUBLIC int rtp_session_unsplice(RtpSession *session, RtpSession *to_session);
+
+ORTP_PUBLIC bool_t ortp_stream_is_ipv6(OrtpStream *os);
+
+/* RtpBundle api */
+typedef struct _RtpBundle RtpBundle;
+
+ORTP_PUBLIC RtpBundle* rtp_bundle_new(void);
+ORTP_PUBLIC void rtp_bundle_delete(RtpBundle *bundle);
+
+ORTP_PUBLIC int rtp_bundle_get_mid_extension_id(RtpBundle *bundle);
+ORTP_PUBLIC void rtp_bundle_set_mid_extension_id(RtpBundle *bundle, int id);
+
+ORTP_PUBLIC void rtp_bundle_add_session(RtpBundle *bundle, const char *mid, RtpSession *session);
+ORTP_PUBLIC void rtp_bundle_remove_session_by_id(RtpBundle *bundle, const char *mid);
+ORTP_PUBLIC void rtp_bundle_remove_session(RtpBundle *bundle, RtpSession *session);
+ORTP_PUBLIC void rtp_bundle_clear(RtpBundle *bundle);
+
+ORTP_PUBLIC RtpSession* rtp_bundle_get_primary_session(RtpBundle *bundle);
+ORTP_PUBLIC void rtp_bundle_set_primary_session(RtpBundle *bundle, const char * mid);
+
+ORTP_PUBLIC const char *rtp_bundle_get_session_mid(RtpBundle *bundle, RtpSession *session);
+
+ORTP_PUBLIC int rtp_bundle_send_through_primary(RtpBundle *bundle, bool_t is_rtp, mblk_t *m, int flags, const struct sockaddr *destaddr, socklen_t destlen);
+/* Returns FALSE if the rtp packet or at least one of the RTCP packet (compound) was for the primary */
+ORTP_PUBLIC bool_t rtp_bundle_dispatch(RtpBundle *bundle, bool_t is_rtp, mblk_t *m);
+ORTP_PUBLIC void rtp_session_use_local_addr(RtpSession * session, const char * rtp_local_addr, const char * rtcp_local_addr);
+
 #ifdef __cplusplus
 }
 #endif
